/* 
 * linux/include/linux/bossa_rts_kernel.h 
 *   Copyright (C) 2004 EMN, DIKU. All rights reserved.
 *   Ecole des Mines de Nantes, University of Copenhagen.
 *
 * This file describes the interface between the RTS and the kernel.
 *
 */

// functions and variables defined by the RTS and used by the kernel
// this is to be included in sched.h

#include "bossa_events.h"

/*******************************************************
 * initialization :
 *
 *     bossa_init, bossa_ioctl_init : initialization functions
 *     bossa_scheduler_requested, bossa_scheduler_name : requesting a
 *       specific scheduler at boot time
 *
 *******************************************************/

extern void bossa_init(void);
extern void bossa_ioctl_init(void);
extern int bossa_scheduler_requested;
extern char *bossa_scheduler_name;

/*******************************************************
 * event notifications (basic scheduling operations) :
 *
 * Create a process      : rts_create_process
 * Terminate a process   : rts_terminate_process
 * Clock tick            : rts_clocktick
 * Unblock a process     : rts_unblock
 * Yield a process       : rts_yield
 * Preempt the current process (with or without blocking) and elect a new
 *    process to the CPU : rts_schedule
 *
 *******************************************************/

extern void rts_create_process_wake(struct task_struct *p);
extern void rts_create_process_sleep(struct task_struct *p);
extern void rts_terminate_process(struct task_struct *p);
extern void rts_clocktick(void);
extern int rts_unblock_preemptive(struct task_struct * p);
extern int rts_unblock_nonpreemptive(struct task_struct * p);
extern void rts_yield(struct task_struct *p);
extern struct task_struct *rts_schedule(struct task_struct *prev);
extern void rts_set_priority(struct task_struct * p,int);
struct ctl_table;
extern int rts_scheduler_info( struct ctl_table * table, int write,
			       struct file * filp, void * buffer,
			       size_t * lenp );
extern int rts_proc_info(struct task_struct *task, char * buffer);

extern int rts_ready(void);
extern int rts_blocked(void);

/*******************************************************
 *
 * CIL still generates code for the specific events
 *
 *******************************************************/

#define schedule_from_idle bossa_schedule
#define schedule_uninterruptible bossa_schedule
#define schedule_uninterruptible_interruptible bossa_schedule
#define schedule_interruptible bossa_schedule
#define schedule_interruptible_pause bossa_schedule
#define schedule_yield_immediate bossa_schedule
#define schedule_yield_pause bossa_schedule
#define schedule_timeout_yield_immediate bossa_schedule_timeout
#define schedule_timeout_uninterruptible bossa_schedule_timeout
#define schedule_timeout_interruptible bossa_schedule_timeout

extern asmlinkage void bossa_schedule(int caller);
extern signed long bossa_schedule_timeout(int caller, signed long timeout);

/*******************************************************
 * bossa_struct : 
 *
 *     used to define Bossa processes
 *     every process scheduler must define a structure for child processes
 *     that begins with this format
 *
 *    *attr points to the default process atributes (pid, memory context)
 *
 *******************************************************/

// seem to have to expose this for the size of task_struct in sched.c

#ifndef BOSSA_STRUCT
#define BOSSA_STRUCT
struct bossa_struct {
  /* default attributes for Bossa */
  /* next and previous processes in the current bossa queue */
  struct list_head list_info;
  /* list of virtual schedulers on the way to this process */
  struct next_list_struct *next_list;
  /* process state */
  int state;
};
#endif
