// To be included in sched.c

inline int
rts_unblock_preemptive(struct task_struct * p) {
  unsigned long flags;

  /* In Bossa, we only notify an unblock event if the process is not
   * running or in the ready queue In Linux we may have useless/multiple
   * calls to wake_up which may cause the Bossa run-time system to notify
   * an unblock event more than once (an unneeded/inexisting unblock event)
   * */

    int res;
#ifndef CONFIG_BOSSA_ALLOW_PREEMPT
#ifdef CONFIG_PREEMPT
  preempt_disable();
#endif
#endif
    spin_lock_irqsave(&bossa_scheduler_lock, flags);
#ifdef CONFIG_BOSSA_TRACE
    /* CA addition */
    if (toggle_trace && t_index < 3000) {
	    do_gettimeofday(& t_buffer[t_index].time);
	    //__asm__ __volatile__ ("RDTSC" : "=A" (t_buffer[t_index].time));
	    //do_div(t_buffer[t_index].time, cpu_khz);
 	    t_buffer[t_index].type = EVENT_UNBLOCK;
 	    t_buffer[t_index].tgt = (p)->pid;
 	    t_buffer[t_index].cur = (current)->pid;
 	    t_index++;
    }
#endif /* CONFIG_BOSSA_TRACE */
    res = root_scheduler.ops->unblock_preemptive_entry(EVENT_UNBLOCK_PREEMPTIVE,(&(p->bossa)));
    spin_unlock_irqrestore(&bossa_scheduler_lock, flags);
#ifndef CONFIG_BOSSA_ALLOW_PREEMPT
#ifdef CONFIG_PREEMPT
  preempt_enable();
#endif
#endif
    return res;
}


inline int
rts_unblock_nonpreemptive(struct task_struct * p) {
  unsigned long flags;

  /* In Bossa, we only notify an unblock event if the process is not
   * running or in the ready queue In Linux we may have useless/multiple
   * calls to wake_up which may cause the Bossa run-time system to notify
   * an unblock event more than once (an unneeded/inexisting unblock event)
   * */

  int res;
#ifndef CONFIG_BOSSA_ALLOW_PREEMPT
#ifdef CONFIG_PREEMPT
  preempt_disable();
#endif
#endif
    spin_lock_irqsave(&bossa_scheduler_lock, flags);
    res = bossa_check_process_state(&(p->bossa),BLOCKED);
    bossa_notify_event(EVENT_UNBLOCK_NONPREEMPTIVE,(&(p->bossa)),NULL);
    spin_unlock_irqrestore(&bossa_scheduler_lock, flags);
#ifndef CONFIG_BOSSA_ALLOW_PREEMPT
#ifdef CONFIG_PREEMPT
  preempt_enable();
#endif
#endif
    if (res && root_s_state == READY) {
      set_tsk_need_resched(current);
    }
    return res;
}


// scheduling functions

#define attr(x) list_entry(x, struct task_struct, bossa)

static inline void yield_system_immediate(void) {
#ifdef CONFIG_BOSSA_TRACE
    /* CA addition */
    if (toggle_trace && t_index < 3000) {
	    do_gettimeofday(& t_buffer[t_index].time);
	    t_buffer[t_index].type = EVENT_YIELD;
 	    t_buffer[t_index].tgt = (current)->pid;
 	    t_buffer[t_index].cur = (current)->pid;
 	    t_index++;
    }
#endif /* CONFIG_BOSSA_TRACE */
  root_s_state =
    root_scheduler.ops->yield_system_immediate_entry(0,running);
}

static inline void block(void) {
#ifdef CONFIG_BOSSA_TRACE
    /* CA addition */
    if (toggle_trace && t_index < 3000) {
	    do_gettimeofday(& t_buffer[t_index].time);
	    t_buffer[t_index].type = EVENT_BLOCK;
	    t_buffer[t_index].tgt = (current)->pid;
	    t_buffer[t_index].cur = (current)->pid;
	    t_index++;
    }
#endif /* CONFIG_BOSSA_TRACE */
  root_s_state = root_scheduler.ops->block_entry(0,running);
}

static inline void call_bossa_schedule(void) {
  root_s_state = root_scheduler.ops->bossa_schedule_entry();
}

inline asmlinkage
struct task_struct *rts_schedule(struct task_struct *prev)
{
	if (running == rts_idle_process) {
	  call_bossa_schedule();
	}
	else {
	  if (prev->state && !(preempt_count() & PREEMPT_ACTIVE)) {
	    if (unlikely((prev->state & TASK_INTERRUPTIBLE) &&
			 unlikely(signal_pending(prev)))) {
	      prev->state = TASK_RUNNING;
	      yield_system_immediate();
	    }
	    else {
	      if (unlikely(prev->exit_state == EXIT_DEAD)) {
		if (root_s_state == READY)
		  call_bossa_schedule();
		else running = rts_idle_process;
	      }
	      else block();
	    }
	  }
	  else { /* state is TASK_RUNNING */
	    switch (root_s_state) {
	    case RUNNING:
	      yield_system_immediate();
	      break;
	    case READY:
	      call_bossa_schedule();
	      break;
	    case BLOCKED:
	    default:
	      running = rts_idle_process;
	      break;
	    }
	  }
	}
#ifdef CONFIG_BOSSA_TRACE
    /* CA addition */
    if (toggle_trace && t_index < 3000) {
	    do_gettimeofday(& t_buffer[t_index].time);
 	    t_buffer[t_index].type = EVENT_BOSSA_SCHEDULE;
 	    t_buffer[t_index].tgt = 0;
 	    t_buffer[t_index].cur = attr(running)->pid;
 	    t_index++;
    }
#endif /* CONFIG_BOSSA_TRACE */
	return attr(running);
}
