#define init_var(x) *(x) = NULL
#define init_queue(x) (x)->next = (x)->prev = x

/* p and queue are always identifiers.  The following are always used as
   expressions. */

#define empty_proc(p) ((p) == NULL)
#define empty_queue list_empty

#define del_from_queue(p) list_del(&((p)->list_info))
#define add_to_head_queue(p,q) list_add(&((p)->list_info),q)
#define add_to_tail_queue(p,q) list_add_tail(&((p)->list_info),q)

/* In the following, the first argument is the process and the second
argument is the destination state.  If either the process or the
destination is a process variable, the operation must be a macro */

#define move_nowhere_queue(p,q) add_to_head_queue(p,q)
#define move_nowhere_lifo(p,q) add_to_head_queue(p,q)
#define move_nowhere_fifo(p,q) add_to_tail_queue(p,q)

#define move_nowhere_proc(p,q) q = p

#define move_queue_nowhere(p,q) del_from_queue(p)

#define move_proc_nowhere(p,q) p = NULL

#define move_proc_queue(p,q) do { add_to_head_queue(p,q); p = NULL; } while(0)

#define move_proc_lifo move_proc_queue

#define move_proc_fifo(p,q) do { add_to_tail_queue(p,q); p = NULL; } while(0)

#define move_queue_proc(p,q) do { PROTOCOL_BOSSA_STRUCT *temp = p; del_from_queue(temp); q = temp; } while(0)

// a macro for uniformity...
#define move_queue_queue(p,q) do {  PROTOCOL_BOSSA_STRUCT *temp = p; del_from_queue(temp);   add_to_head_queue(temp, q); } while(0)

#define move_queue_lifo move_queue_queue

// a macro for uniformity...
#define move_queue_fifo(p,q) do {  PROTOCOL_BOSSA_STRUCT *temp = p;  del_from_queue(temp);   add_to_tail_queue(temp, q); } while(0)

#define move_proc_proc(p,q) do {  q = p; p = NULL; } while(0)

// select from a single queue --------------------------------------------

#ifdef GOODNESS
static inline
PROTOCOL_BOSSA_STRUCT *select_sort(struct list_head *q) {
  // first in the ready queue
  struct list_head *elem;
  struct list_head *q1 = q->next;
  PROTOCOL_BOSSA_STRUCT *chosen = bossa_list_entry(q1);
  int m;
  elem = q1->next;
  if (elem == q) return chosen;
  m = goodness(chosen);
  prefetch(elem->next);
   do {
     PROTOCOL_BOSSA_STRUCT *p = bossa_list_entry(elem);
    int m1 = goodness(p);
    if (m1 > m)//(greater(p, chosen))
      {
        chosen = p;
	m = m1;
      }
      elem = elem->next;
      prefetch(elem->next);
   } while(elem != q);
  return chosen;
}
#else
static inline
PROTOCOL_BOSSA_STRUCT *select_sort(struct list_head *q) {
  // first in the ready queue
  struct list_head *elem;
  struct list_head *q1 = q->next;
  PROTOCOL_BOSSA_STRUCT *chosen = bossa_list_entry(q1);
  elem = q1->next;
  prefetch(elem->next);
  while(elem != q) {
    PROTOCOL_BOSSA_STRUCT *p = bossa_list_entry(elem);
    if (greater(p, chosen)) {
      chosen = p;
    }
    elem = elem->next;
    prefetch(elem->next);
  }
  return chosen;
}
#endif


#define select_lifo_sort select_sort
#define select_fifo_sort select_sort

#define select_sorted(q) bossa_list_entry((q)->next)
#define select_fifo_sorted(q) bossa_list_entry((q)->next)

static inline
int static_greater(PROTOCOL_BOSSA_STRUCT *p1, PROTOCOL_BOSSA_STRUCT *p2) {
  if (static_compare(p1,p2) == 1) return 1;
  return 0;
}

static inline
int static_select_greater(PROTOCOL_BOSSA_STRUCT *p1, PROTOCOL_BOSSA_STRUCT *p2) {
  if (static_select_compare(p1,p2) == 1) return 1;
  return 0;
}

static inline
int static_select_greater_lifo_fifo(PROTOCOL_BOSSA_STRUCT *p1, PROTOCOL_BOSSA_STRUCT *p2) {
  if (static_select_compare_lifo_fifo(p1,p2) == 1) return 1;
  return 0;
}

/* scan as long as the elements are statically equal to the first one */

static inline
PROTOCOL_BOSSA_STRUCT *select_fifo_partially_sorted(struct list_head *q) {
  struct list_head *ptr;
  PROTOCOL_BOSSA_STRUCT *prev = bossa_list_entry(q->next);
  for (ptr = q->next->next, prefetch(ptr->next); ptr != q;
       ptr = ptr->next, prefetch(ptr->next)) {
    PROTOCOL_BOSSA_STRUCT *cur = bossa_list_entry(ptr);
    if (static_compare(prev,cur)) break;
    if (dynamic_select_greater(cur,prev))
      prev = cur;
  }
  return prev;
}

#define select_lifo_partially_sorted select_fifo_partially_sorted
#define select_partially_sorted select_fifo_partially_sorted


// select, where the result is used for comparison with p ----------------

static inline int
select_greater_partially_sorted(PROTOCOL_BOSSA_STRUCT *p,
				struct list_head *q) {
  struct list_head *ptr = q->next;
  /* only test once, because everything in the queue is the same in this
     respect */
  int res = static_select_compare(bossa_list_entry(ptr),p);
  if (res == 1) return 1;
  if (res == -1) return 0;
  prefetch(ptr->next);
  for (; ptr!=q; ptr = ptr->next, prefetch(ptr->next)) {
    int res = static_compare(bossa_list_entry(ptr),p);
    if (res == 1) return 1;
    if (res == -1) return 0;
    if (dynamic_greater(bossa_list_entry(ptr),p)) return 1;
  }
  return 0;
} 

static inline int
select_greater_partially_sorted_lifo_fifo(PROTOCOL_BOSSA_STRUCT *p,
					  struct list_head *q) {
  struct list_head *ptr = q->next;
  /* only test once, because everything in the queue is the same in this
     respect */
  int res = static_select_compare_lifo_fifo(bossa_list_entry(ptr),p);
  if (res == 1) return 1;
  if (res == -1) return 0;
  prefetch(ptr->next);
  for (; ptr!=q; ptr = ptr->next, prefetch(ptr->next)) {
    int res = static_compare_lifo_fifo(bossa_list_entry(ptr),p);
    if (res == 1) return 1;
    if (res == -1) return 0;
    if (dynamic_greater_lifo_fifo(bossa_list_entry(ptr),p)) return 1;
  }
  return 0;
} 

// select, where the result is used for comparison with p
/* not sorted at all */
static inline int
select_greater(PROTOCOL_BOSSA_STRUCT *p, struct list_head *q) {
  struct list_head *ptr = q->next;
  if (static_select_greater(p,bossa_list_entry(ptr))) return 0;
  prefetch(ptr->next);
  for (; ptr!=q; ptr = ptr->next, prefetch(ptr->next)) {
    if (greater(bossa_list_entry(ptr),p)) return 1;
  }
  return 0;
} 

static inline int
select_greater_lifo_fifo(PROTOCOL_BOSSA_STRUCT *p, struct list_head *q) {
  struct list_head *ptr = q->next;
  if (static_select_greater_lifo_fifo(p,bossa_list_entry(ptr))) return 0;
  prefetch(ptr->next);
  for (; ptr!=q; ptr = ptr->next, prefetch(ptr->next)) {
    if (greater_lifo_fifo(bossa_list_entry(ptr),p)) return 1;
  }
  return 0;
} 

// insert ----------------------------------------------------------------

static inline void
insert_psorted_fifo(PROTOCOL_BOSSA_STRUCT *p, struct list_head *q) {
  struct list_head *ptr;
  list_for_each(ptr,q) {
    if (static_greater(p,bossa_list_entry(ptr))) break;
  }
  add_to_head_queue(p,ptr->prev);
}

static inline void
insert_psorted_lifo(PROTOCOL_BOSSA_STRUCT *p, struct list_head *q) {
  struct list_head *ptr;
  list_for_each(ptr,q) {
    if (!static_greater(bossa_list_entry(ptr),p)) break;
  }
  add_to_head_queue(p,ptr->prev);
}

// greater is more efficient than static_greater
static inline void
insert_sorted_fifo(PROTOCOL_BOSSA_STRUCT *p, struct list_head *q) {
  struct list_head *ptr;
  list_for_each(ptr,q) {
    if (greater(p,bossa_list_entry(ptr))) break;
  }
  add_to_head_queue(p,ptr->prev);
}

static inline void
insert_sorted_lifo(PROTOCOL_BOSSA_STRUCT *p, struct list_head *q) {
  struct list_head *ptr;
  list_for_each(ptr,q) {
    if (!greater(bossa_list_entry(ptr),p)) break;
  }
  add_to_head_queue(p,ptr->prev);
}

#define move_nowhere_partially_sorted_fifo(p,q) insert_psorted_fifo(p,q)
#define move_nowhere_sorted_fifo(p,q) insert_sorted_fifo(p,q)

#define move_proc_partially_sorted_fifo(p,q) do { insert_psorted_fifo(p,q); p = NULL; } while(0)

// a macro for uniformity...
#define move_queue_partially_sorted_fifo(p,q) do { del_from_queue(p); insert_psorted_fifo(p,q); } while(0)

#define move_nowhere_partially_sorted_lifo(p,q) insert_psorted_lifo(p,q)
#define move_nowhere_sorted_lifo(p,q) insert_sorted_lifo(p,q)
#define move_nowhere_sorted(p,q) insert_sorted_lifo(p,q)

#define move_proc_partially_sorted_lifo(p,q) do { insert_psorted_lifo(p,q); p = NULL; } while(0)

// a macro for uniformity...
#define move_queue_partially_sorted_lifo(p,q) do {  del_from_queue(p); insert_psorted_lifo(p,q); } while(0)

#define move_proc_sorted_lifo(p,q) do { insert_sorted_lifo(p,q); p = NULL; } while(0)

#define move_proc_sorted_fifo(p,q) do { insert_sorted_fifo(p,q); p = NULL; } while(0)

// a macro for uniformity...
#define move_queue_sorted_lifo(p,q) do {  del_from_queue(p);  insert_sorted_lifo(p,q); } while(0)

#define move_queue_sorted_fifo(p,q) do {  del_from_queue(p);  insert_sorted_fifo(p,q); } while(0)

#define move_queue_sorted move_queue_sorted_lifo
#define move_proc_sorted move_proc_sorted_lifo

#define direct_move_queue_queue(from,to) do { __list_add(to,(from)->prev,(from)->next); INIT_LIST_HEAD(from); } while(0)

/* used by BSD... */

/*static inline int min (int x, int y) {
  if (x < y) return x; else return y;
}

static inline int power (int base, int exp) {
  int acc = 1;
  while (exp--) acc *= base;
  return acc;
}
*/

#ifdef PROCESS_SCHEDULER
static void print_var(char *str, PROTOCOL_PROCESS_STRUCT *p) {
  if (p) {
  printk(NAME ": %s ",str);
  if (p == NULL) printk(KERN_EMERG "NULL\n");
  else printk(KERN_EMERG "%d\n",(attr((struct bossa_struct *)p))->pid);
  }
}

static void print_queue(char *str, struct list_head *q) {
  int x=0;
  struct list_head *elem;
  if (!empty_queue(q)) {
    printk(KERN_EMERG NAME ": %s ",str);
    list_for_each(elem,q) {
      PROTOCOL_PROCESS_STRUCT *p =
	list_entry(elem, PROTOCOL_PROCESS_STRUCT, list_info);
      if (x == 20) {
	x=0;
	printk(KERN_EMERG "\n");
      }
      printk(KERN_EMERG "%d ",(attr((struct bossa_struct *)p))->pid);
      x++;
    }
    printk(KERN_EMERG "\n");
  }
}

static void print_array_queue(char *str, int index, struct list_head *q) {
  printk("%d: ",index);
  print_queue(str,q);
}

static int count_var(PROTOCOL_PROCESS_STRUCT *p) {
  if (p == NULL) return 0; else return 1;
}

static int count_queue(struct list_head *q) {
  struct list_head *p;
  int i=0;
  list_for_each(p,q) i++;
  return i;
}
#endif

#ifdef CONFIG_HIGH_RES_TIMERS
static inline int
normalize_time(unsigned long j, long sj, unsigned long *jiff) {
  while (sj < 0) {
    sj += cycles_per_jiffies;
    (j)--;
  }
  while (sj > cycles_per_jiffies) {
    sj -= cycles_per_jiffies;
    (j)++;
  }
  *jiff = j;
  return sj;
}
#endif

// --------------------------------------------------------------------

/* Functions adapted from Linux 2.6 */

/**
 * __ffs - find first bit in word.
 * @word: The word to search
 *
 * Undefined if no bit exists, so code should check against 0 first.
 */
#ifdef NOT_SUPPORTED
static __inline__ unsigned long __ffs(unsigned long word)
{
	__asm__("bsfl %1,%0"
		:"=r" (word)
		:"rm" (word));
	return word;
}

#ifdef ARRAY_BITMAP
// generic versions

#define bitmap_granularity 32

static inline void inc_bitmap(struct list_head *q, int key) {
  if (empty_queue(&q[key])) {
    int entry = key / bitmap_granularity;
    int offset = key % bitmap_granularity;
    BITMAP[entry] |= 1 << offset;
  }
}

static inline void dec_bitmap(struct list_head *q, int key) {
  if (empty_queue(&q[key])) {
    int entry = key / bitmap_granularity;
    int offset = key % bitmap_granularity;
    BITMAP[entry] &= ~(1 << offset);
  }
}

static void init_bitmap(void) {
  int i;
  for (i=0; i!=bitmap_size; i++) BITMAP[i] = 0;
}

static inline int empty_array(void) {
  int i;
  for (i=0; i!=bitmap_size; i++) if (BITMAP[i]) return 0;
  return 1;
}

static inline struct list_head *first_non_empty(struct list_head *q) {
  int i;
  for (i=0; i!=bitmap_size; i++)
    if (BITMAP[i])
      return &q[__ffs(BITMAP[i]) + (i*bitmap_granularity)];
  panic("first_non_empty: unexpected empty queue");
}
#endif
#endif /* NOT_SUPPORTED */

#ifdef WORD_BITMAP

// versions specialized to the case where there are <= 32 entries

static void init_bitmap(void) {
  BITMAP = 0;
}

static inline void inc_bitmap(struct list_head *q, int key) {
  if (empty_queue(&q[key])) {
    BITMAP |= 1 << key;
  }
}

static inline void dec_bitmap(struct list_head *q, int key) {
  if (empty_queue(&q[key])) {
    BITMAP &= ~(1 << key);
  }
}

static inline int empty_array(void) {
  return !BITMAP;
}

static inline struct list_head *first_non_empty(struct list_head *q) {
  int x = __ffs(BITMAP);
  return &q[x];
}

#endif
